@@ -7,6 +7,7 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);
 require Exporter;
 require DynaLoader;
 require AutoLoader;
+use overload '""' => 'asString';
 
 @ISA = qw(Exporter DynaLoader);
 
@@ -24,7 +25,7 @@ require AutoLoader;
 
 # Optional exports
 @EXPORT_OK = qw(rearrange);
-$VERSION = '1.54';
+$VERSION = '1.62';
 
 require Ace::Object;
 require Ace::Iterator;
@@ -32,7 +33,7 @@ require Ace::Iterator;
 sub AUTOLOAD {
   my $constname;
   ($constname = $AUTOLOAD) =~ s/.*:://;
-  my $val = constant($constname, @_ ? $_[0] : 0);
+  my $val = constant($constname, 0);
   if ($! != 0) {
     if ($! =~ /Invalid/) {
       $AutoLoader::AUTOLOAD = $AUTOLOAD;
@@ -58,27 +59,37 @@ $Ace::ERR = '';
 
 sub connect {
     my $class = shift;
-    my ($host,$port,$path,$program,$objclass,$timeout,$query_timeout) = 
-      rearrange(['HOST','PORT','PATH',
-		 'PROGRAM','CLASS','TIMEOUT',
-		'QUERY_TIMEOUT'],@_);
-    $host ||= 'localhost';
-    $port ||= defined(&ACE_PORT) ? &ACE_PORT : 200001;
-    $timeout = 25 unless defined $timeout;
-
-    my $database;
-
-    # open up a connection to a local database
-    if ($path || $program) {
-      require Ace::Local;
-      $database = Ace::Local->connect(@_);
+    my ($host,$port,$path,$program,$objclass,$timeout,$query_timeout,$database);
+    if (@_ == 1) {  # look for host:port
+      if (($host,$port) = $_[0] =~ /^(?:aceserver:\/\/)?([^:]+):(\d+)$/) {
+	$database = Ace::AceDB->new($host,$port);
+      } else {
+	require Ace::Local;
+	($path = $_[0]) =~ s!^\w+://!!;
+	$database = Ace::Local->connect(-path=>$path);
+      }
     } else {
-      my @p = ($host,$port);
-      push(@p,$query_timeout) if defined $query_timeout;
-#      local($SIG{ALRM}) = sub { die "timed out" } ;
-#      $database = eval "alarm($timeout) if $timeout; my \$d = Ace::AceDB->new(\@p);alarm(0); \$d";
-      $database = Ace::AceDB->new(@p);
+      ($host,$port,$path,$program,$objclass,$timeout,$query_timeout) = 
+	rearrange(['HOST','PORT','PATH',
+		   'PROGRAM','CLASS','TIMEOUT',
+		   'QUERY_TIMEOUT'],@_);
+      $host ||= 'localhost';
+      $port ||= defined(&ACE_PORT) ? &ACE_PORT : 200001;
+      $timeout = 25 unless defined $timeout;
+      
+      # open up a connection to a local database
+      if ($path || $program) {
+	require Ace::Local;
+	$database = Ace::Local->connect(@_);
+      } else {
+	my @p = ($host,$port);
+	push(@p,$query_timeout) if defined $query_timeout;
+	#      local($SIG{ALRM}) = sub { die "timed out" } ;
+	#      $database = eval "alarm($timeout) if $timeout; my \$d = Ace::AceDB->new(\@p);alarm(0); \$d";
+	$database = Ace::AceDB->new(@p);
+      }
     }
+
     unless ($database) {
 	$Ace::ERR = "Couldn't open database";
 	return;
@@ -88,6 +99,7 @@ sub connect {
 		      'database'=> $database,
 		      'host'   => $host,
 		      'port'   => $port,
+		      'path'   => $path,
 		      'class'  => $objclass || 'Ace::Object',
 		      'date_style' => 'java',
 		      'auto_save' => 0,
@@ -118,7 +130,6 @@ sub fetch {
     rearrange(['CLASS',['NAME','PATTERN'],'COUNT','OFFSET','QUERY',
 	       ['FILL','FILLED'],'TOTAL'],@_);
   $offset += 0;
-  $count = wantarray ? -1 : 1 unless defined $count;
   $pattern ||= '*';
   $pattern = Ace::AceDB->freeprotect($pattern);
   if (defined $query) {
@@ -138,6 +149,37 @@ sub fetch {
   return wantarray ? @h : $h[0];
 }
 
+# perform an AQL query
+sub aql {
+  my $self = shift;
+  my $query = shift;
+  my $db = $self->db;
+  my $baseclass = $self->{'class'};
+  my $r = $self->raw_query("aql -j $query");
+  if ($r =~ /(AQL error.*)/) {
+    $self->error($1);
+    return;
+  }
+  my @r;
+  foreach (split "\n",$r) {
+    next if m!^//!;
+    next if m!^\0!;
+    my @objects = map { $baseclass->new(Ace::AceDB->split($_),$self,1)} split "\t";
+    push @r,\@objects;
+  }
+  return @r;
+}
+
+# Return the contents of a keyset.  Pattern matches are allowed, in which case
+# the keysets will be merged.
+sub keyset {
+  my $self = shift;
+  my $pattern = shift;
+  $self->raw_query (qq{find keyset "$pattern"});
+  $self->raw_query (qq{follow});
+  return $self->_list;
+}
+
 #########################################################
 # These functions are for low-level (non OO) access only.
 # This is for low-level access only.
@@ -163,7 +205,9 @@ sub read_object {
     return unless $self->{database};
     my $result;
     while ($self->{database}->status == STATUS_PENDING()) {
-      $result .= $self->{database}->read() ;
+      my $data = $self->{database}->read();
+#      $data =~ s/\0//g;  # get rid of nulls in the buffer
+      $result .= $data;
     }
     return $result;
 }
@@ -184,10 +228,20 @@ sub error {
   return $Ace::ERR;
 }
 
-sub DESTROY { 
+# close the database
+sub close {
   my $self = shift;
   $self->raw_query('save')
     if $self->auto_save;
+  foreach (keys %{$self->{iterators}}) {
+    $self->_unregister_iterator($_);
+  }
+  delete $self->{database};
+}
+
+sub DESTROY { 
+  my $self = shift;
+  $self->close;
 }
 
 
@@ -196,16 +250,22 @@ sub DESTROY {
 sub rearrange {
     my($order,@param) = @_;
     return unless @param;
-    
-    return @param unless (defined($param[0]) && substr($param[0],0,1) eq '-');
+    my %param;
 
-    my $i;
-    for ($i=0;$i<@param;$i+=2) {
+    if (ref $param[0] eq 'HASH') {
+      %param = %{$param[0]};
+    } else {
+      return @param unless (defined($param[0]) && substr($param[0],0,1) eq '-');
+
+      my $i;
+      for ($i=0;$i<@param;$i+=2) {
         $param[$i]=~s/^\-//;     # get rid of initial - if present
         $param[$i]=~tr/a-z/A-Z/; # parameters are upper case
+      }
+
+      %param = @param;                # convert into associative array
     }
     
-    my(%param) = @param;                # convert into associative array
     my(@return_array);
     
     local($^W) = 0;
@@ -239,9 +299,12 @@ sub _query {
 sub _list {
   my $self = shift;
   my ($count,$offset) = @_;
-  ($offset,$count) = (0,-1) unless $count;
   my (@result);
-  my $result = $self->raw_query("list -j -b $offset -c $count");
+  my $query = 'list -j';
+  $query .= " -b $offset" if defined $offset;
+  $query .= " -c $count"  if defined $count;
+  my $result = $self->raw_query($query);
+  $result =~ s/\0//g;  # get rid of &$#&@( nulls
   foreach (split("\n",$result)) {
     next unless my ($class,$name) = Ace::AceDB->split($_);
     push(@result,$self->{'class'}->new($class,$name,$self,1));
@@ -254,9 +317,11 @@ sub _fetch {
   my $self = shift;
   my ($count,$start) = @_;
   my (@result);
-  my $ts = $self->{'timestamps'} ? '-T' : '';
-  ($start,$count) = (0,-1) unless $count;
-  $self->{database}->query("show -j $ts -b $start -c $count");
+  my $query = "show -j";
+  $query .= ' -T' if $self->{timestamps};
+  $query .= " -b $start"  if defined $start;
+  $query .= " -c $count"  if defined $count;
+  $self->{database}->query($query);
   while (my @objects = $self->_fetch_chunk) {
     push (@result,@objects);
   }
@@ -267,11 +332,12 @@ sub _fetch_chunk {
   my $self = shift;
   return unless $self->{database}->status == STATUS_PENDING();
   my $result = $self->{database}->read();
+  $result =~ s/\0//g;  # get rid of &$#&@!! nulls
   my @chunks = split("\n\n",$result);
   my @result;
   foreach (@chunks) {
     next if m!^//!;
-    next unless /\S/;  # f**ing     appears every so often!
+    next unless /\S/;  # occasional empty lines
     push(@result,$self->{'class'}->newFromText($_,$self));
   }
   return @result;
@@ -285,6 +351,14 @@ sub _alert_iterators {
   undef $self->{'active_list'};
 }
 
+sub asString {
+  my $self = shift;
+  return "acedb://$self->{path}" if $self->{'path'};
+  return "aceserver://$self->{host}:$self->{port}" if $self->{'host'};
+  return ref $self;
+}
+
+
 1;
 
 __END__
@@ -487,6 +561,16 @@ details on using this object.
  
     $low_level = $db->db();
 
+=head2 close() Method
+
+You can explicitly close a database by calling its close() method:
+
+   $db->close();
+
+This is not ordinarily necessary because the database will be
+automatically close when it -- and all objects retrieved from it -- go
+out of scope.
+
 =head1 RETRIEVING ACEDB OBJECTS
 
 Once you have established a connection and have an Ace databaes
@@ -590,6 +674,23 @@ If your request is likely to retrieve very many objects, fetch() many
 consume a lot of memory, even if B<-fill> is false.  Consider using
 B<fetch_many()> instead (see below).
 
+=head2 aql() method
+
+    $count   = $db->aql($aql_query);
+    @objects = $db->aql($aql_query);
+
+Ace::aql() will perform an AQL query on the database.  In a scalar
+context it returns the number of rows returned.  In an array context
+it returns a list of rows.  Each row is an anonymous array containing
+the columns returned by the query as an Ace::Object.
+
+If an AQL error is encountered, will return undef or an empty list and
+set Ace->error to the error message.
+
+Note that this routine is not optimized -- there is no iterator
+defined.  All results are returned synchronously, leading to large
+memory consumption for certain queries.
+
 =head2 put() method
 
    $cnt = $db->put($obj1,$obj2,$obj3);
@@ -628,6 +729,25 @@ For example:
 This method can also be used to parse several objects, but only the
 last object successfully parsed will be returned.
 
+=head2 parse_longtext() method
+
+  $object = $db->parse($title,$text);
+
+This will parse the long text (which may contain carriage returns and
+other funny characters) and place it into the database with the given
+title.  In case of a parse error, the undefined value will be returned
+and a (hopefully informative) description of the error will be
+returned by Ace->error(); otherwise, a LongText object will be returned.
+
+For example:
+
+  $author = $db->parse_longtext('A Novel Inhibitory Domain',<<END);
+  We have discovered a novel inhibitory domain that inhibits
+  many classes of proteases, including metallothioproteins.
+  This inhibitory domain appears in three different gene families studied
+  to date...
+  END
+
 =head2 parse_file() method
 
   @objects = $db->parse_file('/path/to/file');
@@ -733,6 +853,15 @@ present the B<-name> and B<-class> arguments will be ignored.
 
 This is an alias for fetch_many().  It is now deprecated.
 
+=head2 keyset() method
+
+    @objects = $db->keyset($keyset_name);
+
+This method returns all objects in a named keyset.  Wildcard
+characters are accepted, in which case all keysets that match the
+pattern will be retrieved and merged into a single list of unique
+objects.
+
 =head2 grep() method
 
     @objects = $db->grep($grep_string);
@@ -998,11 +1127,12 @@ current tag's entire column.  It returns the current subtree instead.
 
 7. When timestamps are active, many optimizations are disabled. 
 
-8. Item number eight is missing.
+8. Item number eight is still missing.
 
 =head1 SEE ALSO
 
-Jade documentation, L<Ace::Object>, L<Ace::Local>, L<Ace::Model>.
+L<Ace::Object>, L<Ace::Local>, L<Ace::Model>,
+L<Ace::Sequence>,L<Ace::Sequence::Multi>.
 
 =head1 AUTHOR
 
@@ -1083,6 +1213,26 @@ sub parse {
   return $results[0];
 }
 
+# Parse a single object as longtext and return the result
+# as an object
+sub parse_longtext {
+  my $self  = shift;
+  my ($title,$body) = @_;
+      my $mm = "parse =
+Longtext $title
+$body
+***LongTextEnd***
+" ;
+  $mm =~ s/\//\\\//g ;
+  $mm =~ s/\n/\\n/g ;
+  $mm .= "\n" ;
+  my $result = $self->raw_query($mm) ;
+  $Ace::ERR = $result=~/sorry|parse error/mi ? $result : '';
+  my @results = $self->_list(1,0);
+  return $results[0];
+}
+
+
 # Parse a file and return all the results
 sub parse_file {
   my $self = shift;
@@ -1179,7 +1329,6 @@ sub find {
   my ($query,$count,$offset,$filled,$total) = rearrange(['QUERY','COUNT',
 							 'OFFSET',['FILL','FILLED'],'TOTAL'],@_);
   $offset += 0;
-  $count = wantarray ? -1 : 1 unless defined $count;
   $query = "find $query" unless $query=~/^find/i;
   my $cnt = $self->count(-query=>$query);
   $$total = $cnt if defined $total;
@@ -1233,7 +1382,6 @@ sub grep {
   my ($pattern,$count,$offset,$filled,$total,$long) = 
       rearrange(['PATTERN','COUNT','OFFSET',['FILL','FILLED'],'TOTAL','LONG'],@_);
   $offset += 0;
-  $count = wantarray ? -1 : 1 unless defined $count;
   my $grep = defined($long) && $long ? 'LongGrep' : 'grep';
   my $r = $self->raw_query("$grep $pattern");
   my ($cnt) = $r =~ /Found (\d+) objects/m;
